/* Copyright (C) 1998 Odysseas Pentakalos and Aram Khalili

  This program is freely distributable under the GNU General
  Public License Version 2 or later. The GNU GPL is available
  from the Free Software Foundation, Inc., 59 Temple Place,
  Suite 330, Boston, MA 02111-1307, USA.

  The authors can be contacted by email :
  Odysseas Pentakalos : odysseas@cs.umbc.edu
  Aram Khalili        : akhali1@gl.umbc.edu
*/

#include <stdio.h>
#include <string.h>
#include <scsi/sg.h>
#include <scsi/scsi.h>
#include <sys/types.h>
#include "tapelib.h"

int read_elt_stat(int fd, 
		  elt_stat_req_t *esr, 
		  elt_stat_data_t *esd,
		  elt_stat_page_t **esp, 
		  stor_elt_desc_t **sed,
		  data_transf_elt_desc_t **dted,
		  med_transp_elt_desc_t **mted)
{
   typedef struct {
    struct sg_header head;
    u_char cmd[12];
  } SCSI_Cmd_t;

  SCSI_Cmd_t Cmd;
  int i, j, descs, ret, types;
  __u16 word;
  u_int bytes, bytes_left, voltag=0;
  u_char *buf, *ptrBuf, tempBuf[100];
  void print_command(u_char *cmd, int bytes);

  /*
    Prepare the element status command
    */
  for (i=0; i<12; i++)
    Cmd.cmd[i]=0;

  Cmd.cmd[0]=READ_ELEMENT_STATUS;

  /* Byte 01, Bits 7 through 5 are the lun and should be zero 
     Byte 01, Bit 4 is the VolTag bit - 0 means do not return bar code
     labels, and 1 means do return them
     Byte 01, Bits 3 through 0 indicate the element type
              0h - all element types
	      1h - medium transport element ( robot arm )
	      2h - storage element ( cartridge slots )
	      4h - data transfer element ( tape drive ) 
      */
  if (esr->voltag == 1) {
    Cmd.cmd[1] |= 0x10;
    voltag=VOL_TAG;
  }
  
  Cmd.cmd[1] |= esr->type;
  
#ifdef DEBUG
  printf("read_elt_stat: Using a voltag of %d and type of %d set byte 1 to %d\n",
	 esr->voltag, esr->type, Cmd.cmd[1]);
#endif DEBUG

  /* 
     if the user knows the configuration of the library then
     lets compute expected information based on request
     */
  if (esr->initflag==1) {
    
    Cmd.cmd[2] = TWOBYTEMSB(esr->start);
    Cmd.cmd[3] = TWOBYTELSB(esr->start);
    
    /*
      Here we need to compute the maximum amount of data that we expect
      to get back from the tape library. The format of the data stream
      returning is as follows:
      ---------------------
      |  ELT_STAT_DATA    |
      |-------------------|
      |  ELT_STAT_PAGE    | -
      |-------------------|  |  This is there for each of the element
      |  ELT_DESCRIPTOR   |  |  groups that were requested by the command
      |-------------------| -   and there is one element descriptor for
      |        ...        |     each element of the particular type
      |-------------------|
      |  ELT_DESCRIPTOR   |     The size of each descriptor varies based
      |-------------------|     on whether the user requested the volume
                                tag information. If voltag==1 then the size
				of the descriptor increases by VOL_TAG
				bytes.

    Also, specifically for the data transfer element, if a serial number was 
    requested (snreq==1) then the size of the descriptor increases further 
    by TAPE_DRIVE_SN bytes
    */

    word=0;
    bytes=ELT_STAT_DATA;

    if (esr->type == ALL_ELT_TYPES | esr->type == MED_TRANSP_ELT_TYPE) {
      word += esr->num_med_transp_elt;
      bytes += ELT_STAT_PAGE + 
	(MED_TRANSP_ELT+voltag)*esr->num_med_transp_elt;
    }

    if (esr->type == ALL_ELT_TYPES | esr->type == STOR_ELT_TYPE) {
      word += esr->num_stor_elt;
      bytes += ELT_STAT_PAGE + 
	(STOR_ELT+voltag)*esr->num_stor_elt;
    }

    if (esr->type == ALL_ELT_TYPES | esr->type == DATA_TRANSF_ELT_TYPE) {
      word += esr->num_data_transf_elt;
      bytes += ELT_STAT_PAGE + 
	(DATA_TRANSF_ELT+voltag)*esr->num_data_transf_elt; 
      if (esr->snreq)
	bytes += TAPE_DRIVE_SN * esr->num_data_transf_elt;
    }

    /*
      Compute the total number of elements that we want to
      get
      */
    word=esr->num_med_transp_elt+esr->num_stor_elt+esr->num_data_transf_elt;
    Cmd.cmd[4] = TWOBYTEMSB(word);
    Cmd.cmd[5] = TWOBYTELSB(word);

    /*
      Compute the total number of types of elements that we
      want to get (computed based on the type code pased)
      */
    if (esr->type == 0)
      types=3;
    else
      types=1;

    Cmd.cmd[6] = 0;

    /*
      Store the computed maximum number of bytes that we expect to
      get back
      */
    Cmd.cmd[7] = THREEBYTEMSB(bytes);
    Cmd.cmd[8] = THREEBYTEMID(bytes);
    Cmd.cmd[9] = THREEBYTELSB(bytes);

#ifdef DEBUG
    print_command(Cmd.cmd, 10);
#endif DEBUG

    Cmd.cmd[10] = 0;
    Cmd.cmd[11] = 0;

    Cmd.head.pack_len=sizeof(Cmd);
    Cmd.head.reply_len=bytes + sizeof(struct sg_header);

    ret=write(fd, &Cmd, sizeof(Cmd));

#ifdef DEBUG
    printf("read_elt_stat: The write returned %d bytes\n", ret);
#endif

    /*
      Now lets read back the results of the write command
      */
    buf=(u_char *) malloc((bytes + sizeof(struct sg_header)) * sizeof(u_char));
    ret=read(fd, buf, Cmd.head.reply_len);

#ifdef DEBUG
    printf("read_elt_stat: Expected %d bytes and got %d\n",
	   Cmd.head.reply_len, ret);
#endif

    /*
      Skip the struct sg_header header
      */
    buf = buf + sizeof(struct sg_header);

    /*
      Extract the element status data structure
      */
    get_elt_stat_data(buf, esd);

    /*
      Allocate space for the array of element page structures for
      each of the element types requested
      */

    *esp=(elt_stat_page_t *) malloc(types * sizeof(elt_stat_page_t));

    /*
      Skip the element status page data
      */
    buf = buf + ELT_STAT_DATA;
    
    bytes_left=esd->elt_stat_page_bytes;

    i=0;

    while (bytes_left > 0) {

    /*
      Extract the element status page structure
      */

      get_elt_stat_page(buf, &((*esp)[i]));
      
      /*
	Skip this to go to the descriptors
	*/
      buf = buf + ELT_STAT_PAGE;
      bytes_left -= ELT_STAT_PAGE;

      switch((*esp)[i].elt_type) {
      case STOR_ELT_TYPE: 
	*sed=(stor_elt_desc_t *) malloc( (*esp)[i].elt_desc_cnt * 
					 sizeof(stor_elt_desc_t));
	break;
      case MED_TRANSP_ELT_TYPE:
	*mted=(med_transp_elt_desc_t *) malloc( (*esp)[i].elt_desc_cnt *
					 sizeof(med_transp_elt_desc_t));
	break;
      case DATA_TRANSF_ELT_TYPE:
	*dted=(data_transf_elt_desc_t *) malloc( (*esp)[i].elt_desc_cnt *
					 sizeof(data_transf_elt_desc_t));
	break;
      default:
	return -1;

      }

      for (j=0; j < (*esp)[i].elt_desc_cnt; j++) {
	switch ((*esp)[i].elt_type) {
	case STOR_ELT_TYPE:
	  get_stor_elt_desc(buf, &((*sed)[j]));
	  print_stor_elt_desc(&((*sed)[j]));
	  break;
	case MED_TRANSP_ELT_TYPE:
	  get_med_transp_elt_desc(buf, &((*mted)[j]));
	  break;
	case DATA_TRANSF_ELT_TYPE:
	  get_data_transf_elt_desc(buf, &((*dted)[j]));
	  /*print_data_transf_elt_desc(&((*dted)[j]));*/
	  break;
	}
	bytes_left -= (*esp)[i].elt_desc_len;
	buf = buf + (*esp)[i].elt_desc_len;
      } 
      return(1);
    }
  }
}

void print_command(u_char *cmd, int bytes)
{
  int i;

  for (i=0; i < bytes; i++)
    printf("Cmd[%d]=%u ", i, cmd[i]);
  printf("\n");

  return;
} 

int get_elt_stat_data(u_char *buf, elt_stat_data_t *esd)
{
  int UChar_to_UInt(u_int *value, u_char *array, int bytes);

  if (UChar_to_UInt(&(esd->first_elt), &(buf[0]), 2) < 0)
    return -1;

  if (UChar_to_UInt(&(esd->num_elt), &(buf[2]), 2) < 0)
    return -1;

  if (UChar_to_UInt(&(esd->elt_stat_page_bytes), &(buf[5]), 3) < 0)
    return -1;
  
  return 1;
}
 
int get_elt_stat_page(u_char *buf, elt_stat_page_t *esp)
{
  esp->elt_type=(__u8) buf[0];
  
  esp->pvoltag_flag=(buf[0] & 0x80) ? 1 : 0;

  if (UChar_to_UInt(&(esp->elt_desc_len), &(buf[2]), 2) < 0)
    return -1;
  
  if (UChar_to_UInt(&(esp->elt_desc_len_tot), &(buf[5]), 3) < 0)
    return -1;

  if (esp->elt_desc_len > 0)
    esp->elt_desc_cnt=esp->elt_desc_len_tot/esp->elt_desc_len;

  return 1;
}

int get_stor_elt_desc(u_char *buf, stor_elt_desc_t *sed) 
{
  if (UChar_to_UInt(&(sed->elt_addr), &(buf[0]), 2) < 0)
    return -1;

  sed->access=(buf[2] & 0x08) ? 1 : 0;
  sed->except=(buf[2] & 0x04) ? 1 : 0;
  sed->full=(buf[2] & 0x01) ? 1 : 0;

  sed->add_sense= (__u8) buf[4];
  sed->add_sense_qual= (__u8) buf[5];

  sed->svalid= (buf[9] & 0x80) ? 1 : 0;
  sed->invert= (buf[9] & 0x40) ? 1 : 0;

  if (UChar_to_UInt(&(sed->source_elt_addr), &(buf[10]), 2) < 0)
    return -1;

  return 1;
}

void print_stor_elt_desc(stor_elt_desc_t *sed)
{
  printf("Element at address %d\n", sed->elt_addr);
  
  if (sed->access==1)
    printf("\tThe cartridge magazine is accessible\n");
  else
    printf("\tThe cartridge magazine is not accessible\n");

  if (sed->except==1)
    printf("\tThe cartridge is in an exceptional state\n");
  else
    printf("\tThe cartridge is in a normal state\n");

  if (sed->full==1)
    printf("\tThe cartridge slot contains a cartridge\n");
  else
    printf("\tThe cartridge slot does not contain a cartridge\n");

  printf("\tThe additional sense code is %x (h)\n",
	 sed->add_sense);
  printf("\tThe additional sense code qualifier is %x (h)\n", 
	 sed->add_sense_qual);

  if (sed->svalid==1)
    printf("\tSource Element address is %d\n", sed->source_elt_addr);

  printf("\n");


  return;
}

int get_med_transp_elt_desc(u_char *buf, med_transp_elt_desc_t *mted)
{
  return 1;
}

int get_data_transf_elt_desc(u_char *buf, data_transf_elt_desc_t *dted)
{
  return 1;
}
  
void print_elt_stat_page(elt_stat_page_t *esp)
{
  printf("Element Type Code = %u\n", esp->elt_type);
  printf("Element Descriptor Length = %u\n", esp->elt_desc_len);
  printf("Total Bytes of Element Descriptor Data = %u\n",
	 esp->elt_desc_len_tot);
  printf("Number of Element Descriptors = %u\n", esp->elt_desc_cnt);

  return;
}

void print_elt_stat_data(elt_stat_data_t *esd)
{
  printf("First Element Address Reported = %u\n", esd->first_elt);
  printf("Number of Elements = %u\n", esd->num_elt);
  printf("Total number of Element Status Page bytes = %u\n", 
	 esd->elt_stat_page_bytes);

  return;
}

int UChar_to_UInt(u_int *value, u_char *array, int bytes)
{
  int i;

  *value=0;

  for (i=0;  i < bytes; i++)
    if (array+i != NULL)
      *value += array[i] << 8*(bytes-i-1);
    else
      return -1;

  return *value;
}










